from threading import Event
from guizero import App, ButtonGroup
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    assert b.master == a
    assert b.value == "foo"
    assert b.value_text == "foo"
    assert b.grid == None
    assert b.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    b = ButtonGroup(
        a,
        ["foo", "bar"],
        "bar",
        grid = [0,1],
        align = "top",
        width=11,
        height=10)

    assert b.value == "bar"
    assert b.value_text == "bar"
    assert b.grid[0] == 0
    assert b.grid[1] == 1
    assert b.align == "top"
    assert b.width == 11
    assert b.height == 10
    a.destroy()

def test_2d_options_list():
    a = App()
    b = ButtonGroup(a, [["foo", "f"], ["bar", "b"]])
    assert b.value == "f"
    assert b.value_text == "foo"
    a.destroy()

def test_getters_setters():
    a = App()
    b = ButtonGroup(a, ["foo", ["bar", "b"]])
    assert b.value == "foo"
    assert b.value_text == "foo"

    b.value = "b"
    assert b.value == "b"
    assert b.value_text == "bar"

    b.value_text = "foo"
    assert b.value == "foo"
    assert b.value_text == "foo"

    a.destroy()

def test_append():
    a = App()
    b = ButtonGroup(a, [["foo", "f"], ["bar", "b"]])

    assert b.options == [["foo", "f"], ["bar", "b"]]

    b.append("car")
    assert b.options == [["foo", "f"], ["bar", "b"], ["car", "car"]]

    b.append(["lah", "l"])
    assert b.options == [["foo", "f"], ["bar", "b"], ["car", "car"], ["lah", "l"]]

    a.destroy()

def test_insert():
    a = App()
    b = ButtonGroup(a, [["foo", "f"], ["bar", "b"]])

    assert b.options == [["foo", "f"], ["bar", "b"]]

    b.insert(1, "car")
    assert b.options == [["foo", "f"], ["car", "car"], ["bar", "b"]]

    b.insert(2, ["lah", "l"])
    assert b.options == [["foo", "f"], ["car", "car"], ["lah", "l"], ["bar", "b"]]

    a.destroy()

def test_remove():
    a = App()
    b = ButtonGroup(a, [["foo", "f"], ["bar", "b"], ["car", "c"]])

    assert b.options == [["foo", "f"], ["bar", "b"], ["car", "c"]]
    b.remove("f")
    assert b.options == [["bar", "b"], ["car", "c"]]

    a.destroy()

def test_clear():
    a = App()
    b = ButtonGroup(a, [["foo", "f"], ["bar", "b"]])

    b.clear()
    assert len(b.options) == 0
    assert b.value == ""

    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        assert b.value == "bar"
        assert b.value_text == "bar"
        callback_event.set()

    b = ButtonGroup(a, ["foo", "bar"], command = callback)

    assert not callback_event.is_set()
    b._rbuttons[1].tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_command_with_args():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "foo"
        assert b.value == "bar"
        assert b.value_text == "bar"
        callback_event.set()

    b = ButtonGroup(a, ["foo", "bar"], command = callback, args = ["foo"])

    b._rbuttons[1].tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    b = ButtonGroup(a, ["foo", "bar"])

    b._rbuttons[1].tk.invoke()
    assert not callback_event.is_set()

    b.update_command(callback)
    b._rbuttons[1].tk.invoke()
    assert callback_event.is_set()
    callback_event.clear()

    b.update_command(None)
    b._rbuttons[1].tk.invoke()
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_args():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "foo"
        callback_event.set()

    b = ButtonGroup(a, ["foo", "bar"])

    b.update_command(callback, ["foo"])
    b._rbuttons[1].tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_after_schedule():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    schedule_after_test(a, b)
    a.destroy()

def test_repeat_schedule():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    schedule_repeat_test(a, b)
    a.destroy()

def test_destroy():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    destroy_test(b)
    a.destroy()

def test_enable():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    enable_test(b)
    a.destroy()

def test_display():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    display_test(b)
    a.destroy()

def test_text():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    text_test(b)
    a.destroy()

def test_color():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    color_test(b)
    a.destroy()

def test_size():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    size_text_test(b)
    size_fill_test(b)
    a.destroy()

def test_events():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    events_test(b)
    a.destroy()

def test_cascaded_properties():
    a = App()
    b = ButtonGroup(a, ["foo", "bar"])
    cascaded_properties_test(a, b, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: ButtonGroup(a, ["foo", "bar"]), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = ButtonGroup(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = ButtonGroup(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = ButtonGroup(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = ButtonGroup(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()



from threading import Event
from guizero import App, Combo
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    c = Combo(a, ["foo", "bar"])
    assert c.master == a
    assert c.value == "foo"
    assert c.grid == None
    assert c.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    c = Combo(
        a,
        ["foo", "bar"],
        selected = "bar",
        grid = [0,1],
        align = "top",
        width=10,
        height=11)

    assert c.value == "bar"
    assert c.grid[0] == 0
    assert c.grid[1] == 1
    assert c.align == "top"
    assert c.width == 10
    assert c.height == 11
    a.destroy()

def test_no_options():
    a = App()
    c = Combo(a)

    assert c.value == ""
    assert len(c.options) == 0

    c.append("foo")
    assert c.value == "foo"

    a.destroy()

def test_getters_setters():
    a = App()
    c = Combo(a, ["foo", "bar"])

    assert c.value == "foo"
    c.value = "bar"
    assert c.value == "bar"

    a.destroy()

def test_select_default():
    a = App()
    c1 = Combo(a, ["foo", "bar"], selected="bar")
    c2 = Combo(a, ["foo", "bar"])

    assert c1.value == "bar"
    c1.value = "foo"
    assert c1.value == "foo"
    c1.select_default()
    assert c1.value == "bar"

    assert c2.value == "foo"
    c2.value = "bar"
    assert c2.value == "bar"
    c2.select_default()
    assert c2.value == "foo"

    a.destroy()

def test_append():
    a = App()
    c = Combo(a, ["foo", "bar"])

    assert c.options == ["foo", "bar"]
    c.append("car")
    assert c.options == ["foo", "bar", "car"]

    a.destroy()

def test_insert():
    a = App()
    c = Combo(a, ["foo", "bar"])

    assert c.options == ["foo", "bar"]
    c.insert(1, "car")
    assert c.options == ["foo", "car", "bar"]

    a.destroy()

def test_remove():
    a = App()
    c = Combo(a, ["foo", "bar", "foo"])

    assert c.options == ["foo", "bar", "foo"]
    c.remove("foo")
    assert c.options == ["bar", "foo"]

    a.destroy()

def test_clear():
    a = App()
    c = Combo(a, ["foo", "bar"])

    c.clear()
    assert len(c.options) == 0
    assert c.value == ""

    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    c = Combo(a, ["foo", "bar"], command = callback)
    assert not callback_event.is_set()
    # you cant invoke a tk optionmenu - this is better than no tests!
    c._command_callback(c.value)
    assert callback_event.is_set()

    a.destroy()

def test_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "foo"
        callback_event.set()

    c = Combo(a, ["foo", "bar"], command = callback)
    assert not callback_event.is_set()

    c._command_callback(c.value)
    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    c = Combo(a, ["foo", "bar"])

    c._command_callback(c.value)
    assert not callback_event.is_set()

    c.update_command(callback)
    c._command_callback(c.value)
    assert callback_event.is_set()
    callback_event.clear()

    c.update_command(None)
    c._command_callback(c.value)
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert c.value == "foo"
        callback_event.set()

    c = Combo(a, ["foo", "bar"])

    c.update_command(callback)

    c._command_callback(c.value)
    assert callback_event.is_set()

    a.destroy()

def test_after_schedule():
    a = App()
    c = Combo(a, ["foo", "bar"])
    schedule_after_test(a, c)
    a.destroy()

def test_repeat_schedule():
    a = App()
    c = Combo(a, ["foo", "bar"])
    schedule_repeat_test(a, c)
    a.destroy()

def test_destroy():
    a = App()
    c = Combo(a, ["foo", "bar"])
    destroy_test(c)
    a.destroy()

def test_enable():
    a = App()
    c = Combo(a, ["foo", "bar"])
    enable_test(c)
    a.destroy()

def test_display():
    a = App()
    c = Combo(a, ["foo", "bar"])
    display_test(c)
    a.destroy()

def test_text():
    a = App()
    c = Combo(a, ["foo", "bar"])
    text_test(c)
    a.destroy()

def test_color():
    a = App()
    c = Combo(a, ["foo", "bar"])
    color_test(c)
    a.destroy()

def test_size():
    a = App()
    c = Combo(a, ["foo", "bar"])
    size_text_test(c)
    size_fill_test(c)
    a.destroy()

def test_events():
    a = App()
    c = Combo(a, ["foo", "bar"])
    events_test(c)
    a.destroy()

def test_cascaded_properties():
    a = App()
    c = Combo(a, ["foo", "bar"])
    cascaded_properties_test(a, c, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: Combo(a, ["foo", "bar"]), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Combo(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Combo(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Combo(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Combo(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()



__name__ = "guizero"
__package__ = "guizero"
__version__ = '1.1.1'
__author__ = "Laura Sach"

from sys import exit
try:
    from tkinter import Tk
except ImportError:
    print("tkinter did not import successfully. Please check your setup.")
    exit(1)

from . import utilities as utils
from .utilities import system_config
from .dialog import *

from .App import App
from .Box import Box
from .ButtonGroup import ButtonGroup
from .CheckBox import CheckBox
from .Combo import Combo
from .Drawing import Drawing
from .ListBox import ListBox
from .MenuBar import MenuBar
from .Picture import Picture
from .PushButton import PushButton
from .RadioButton import RadioButton
from .Slider import Slider
from .Text import Text
from .TextBox import TextBox
from .PushButton import PushButton
from .Waffle import Waffle
from .Window import Window



from threading import Event
from guizero import App, Slider
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_pixel_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    s = Slider(a)
    assert s.tk.cget("from") == 0
    assert s.tk.cget("to") == 100
    assert s.tk.cget("orient") == "horizontal"
    assert s.master == a
    assert s.grid == None
    assert s.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    s = Slider(
        a,
        start=10,
        end=20,
        horizontal=False,
        grid = [0,1],
        align = "top",
        width=10,
        height=11)

    assert s.tk.cget("from") == 10
    assert s.tk.cget("to") == 20
    assert s.tk.cget("orient") == "vertical"
    assert s.grid[0] == 0
    assert s.grid[1] == 1
    assert s.align == "top"
    assert s.width == 10
    assert s.height == 11
    a.destroy()

def test_getters_setters():
    a = App()
    s = Slider(a)

    assert s.value == 0
    s.value = 10
    assert s.value == 10

    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    s = Slider(a, command = callback)
    assert not callback_event.is_set()
    # you cant invoke a tk scale - this is better than no tests!
    s._command_callback(s.value)
    assert callback_event.is_set()

    a.destroy()

def test_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == 0
        callback_event.set()

    s = Slider(a, command = callback)
    assert not callback_event.is_set()

    s._command_callback(s.value)
    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    s = Slider(a)

    s._command_callback(s.value)
    assert not callback_event.is_set()

    s.update_command(callback)
    s._command_callback(s.value)
    assert callback_event.is_set()
    callback_event.clear()

    s.update_command(None)
    s._command_callback(s.value)
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_parameter():
    a = App()

    callback_event = Event()
    def callback(value):
        assert s.value == 0
        callback_event.set()

    s = Slider(a)

    s.update_command(callback)

    s._command_callback(s.value)
    assert callback_event.is_set()

    a.destroy()

def test_after_schedule():
    a = App()
    s = Slider(a)
    schedule_after_test(a, s)
    a.destroy()

def test_repeat_schedule():
    a = App()
    s = Slider(a)
    schedule_repeat_test(a, s)
    a.destroy()

def test_destroy():
    a = App()
    s = Slider(a)
    destroy_test(s)
    a.destroy()

def test_enable():
    a = App()
    s = Slider(a)
    enable_test(s)
    a.destroy()

def test_display():
    a = App()
    s = Slider(a)
    display_test(s)
    a.destroy()

def test_text():
    a = App()
    s = Slider(a)
    text_test(s)
    a.destroy()

def test_color():
    a = App()
    s = Slider(a)
    color_test(s)
    a.destroy()

def test_size():
    a = App()
    s = Slider(a)
    size_pixel_test(s)
    size_fill_test(s)
    a.destroy()

def test_events():
    a = App()
    s = Slider(a)
    events_test(s)
    a.destroy()

def test_cascaded_properties():
    a = App()
    s = Slider(a)
    cascaded_properties_test(a, s, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: Slider(a), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = Slider(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = Slider(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = Slider(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = Slider(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


# this is to cater for Python 2, is it really needed?
try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec

try:
    from PIL import Image, ImageTk
    PIL_AVAILABLE = True
except ImportError:
    PIL_AVAILABLE = False

from tkinter import PhotoImage, TclError
from collections.abc import MutableSequence

import sys

class GUIZeroException(Exception):
    pass

# holds details about the configuration guizero is using
class SystemConfig():
    def __init__(self):
        """
        Holds details about the system configuration guizero is using
        """
        self._platform = sys.platform
        if self._platform.startswith("linux"):
            self._platform = "linux"

        if PIL_AVAILABLE:
            self._supported_image_types = ["GIF", "Animated GIF", "BMP", "ICO", "PNG", "JPG", "TIF"]
        else:
            self._supported_image_types = ["GIF", "PNG"]
            if self._platform == "darwin":
                # macOS only supports GIF with PIL
                self._supported_image_types = ["GIF"]

        # tk options  
        self._tk_options = {
            "*Label.Font": "helvetica 12",
            "*Label.Foreground": "black",
        }

    @property
    def PIL_available(self):
        """
        Returns `True` if PIL (Python Imaging Library) is available.
        """
        return PIL_AVAILABLE

    @property
    def supported_image_types(self):
        """
        Returns a list of images types supported by this system
        e.g. ('GIF', 'PNG', 'JPG')
        """
        return self._supported_image_types

    @property
    def platform(self):
        """
        Returns the current platform ("linux", "darwin", "win32")
        """
        return self._platform

    @property
    def tk_options(self):
        """
        Returns a dictionary of tk options in the format {"pattern": value}
        which will be applied when an App is created.

        The tk options can be used to modify the default behaviour of 
        tk and its widgets e.g. Change the background colour of all Buttons ::

            from guizero import system_config
            system_config.tk_options["*Button.Background"] = "green"

        """
        return self._tk_options

system_config = SystemConfig()


class GUIZeroImage():
    def __init__(self, image_source, width, height):
        """
        GUIZeroImage manages an "image" for guizero widgets, parsing its
        contents, sizing it accordingly and managing environment.

        :param string image_source:
            The source of the image, a file path, PIL or
            Tk image object.

        :param int width:
            The required width of the image, set to `None`, to keep the
            original image width

        :param int height:
            The required height of the image, set to `None`, to keep the
            original image width.
        """
        self._image_source = image_source
        self._pil_image = None
        self._tk_image = None
        self._tk_frames = []
        self._width = width
        self._height = height
        self._current_frame = 0
        self._animation = False
        self._animation_running = False

        # open the image
        self._setup_image()

    @property
    def image_source(self):
        """
        Returns the original source of the image, be that a file path, PIL or
        Tk image object.
        """
        return self._image_source

    @property
    def tk_image(self):
        """
        Returns the Tk PhotoImage object.
        """
        return self._tk_image

    @property
    def pil_image(self):
        """
        Returns the PIL Image object.
        """
        return self._pil_image

    @property
    def width(self):
        """
        Returns the image width.
        """
        return int(self._width)

    @property
    def height(self):
        """
        Returns the image height.
        """
        return int(self._height)

    @property
    def animation(self):
        """
        Returns `True` if the image contains more than 1 frame (i.e. is an
        animation)
        """
        return self._animation

    @property
    def tk_frames(self):
        """
        Returns a list of frames as Tk PhotoImage objects which make up this
        image.
        """
        return self._tk_frames

    def _setup_image(self):
        try:
            # open image
            self._open_image_source()

            # size image
            self._size_image()

            # open frames
            self._open_image_frames()

        except Exception as e:
            error_text = "Image import error - '{}'\n".format(e)
            error_text += "Check the file path and image type is {}".format("/".join(system_config.supported_image_types))
            raise_error(error_text)

    def _open_image_source(self):
        if system_config.PIL_available:
            if isinstance(self._image_source, str):
                # the source is a string, so try and open as a path
                self._pil_image = Image.open(self._image_source)
                self._tk_image = ImageTk.PhotoImage(self._pil_image)

            elif Image.isImageType(self._image_source):
                # the source is a PIL Image
                self._pil_image = self._image_source
                self._tk_image = ImageTk.PhotoImage(self._pil_image)

            elif isinstance(self._image_source, (PhotoImage, ImageTk.PhotoImage)):
                self._tk_image = self._image_source

            else:
                raise Exception("Image must be a file path, PIL.Image or tkinter.PhotoImage")

        else:
            if isinstance(self._image_source, str):
                self._tk_image = PhotoImage(file=self._image_source)

            elif isinstance(self._image_source, PhotoImage):
                self._tk_image = self._image_source

            else:
                raise Exception("Image must be a file path or tkinter.PhotoImage")

    def _size_image(self):

        # if there is no size, set it to the image width
        if self._width is None:
            self._width = self._tk_image.width()

        if self._height is None:
            self._height = self._tk_image.height()

        # does it need resizing?
        if self._width != self._tk_image.width() or self._height != self._tk_image.height():
            if self._pil_image:
                resized_image = self._pil_image.resize((self._width, self._height), Image.ANTIALIAS)
                self._tk_image = ImageTk.PhotoImage(resized_image)
            else:
                error_format("Image resizing - cannot scale the image as PIL is not available.")

    def _open_image_frames(self):
        if self._pil_image:
            frame_count = 0
            try:
                while True:
                    self._pil_image.seek(frame_count)
                    tk_frame = ImageTk.PhotoImage(self._pil_image.resize((self._width, self._height), Image.ANTIALIAS))

                    try:
                        delay = self._pil_image.info['duration']
                    except:
                        delay = 100

                    self._tk_frames.append((tk_frame, delay))
                    frame_count += 1

            except EOFError as e:
                # end of frames
                pass

            if frame_count > 1:
                self._animation = True


class AnimationPlayer():
    def __init__(self, master, guizero_image, update_image_callback):
        """
        AnimationPlayer manages the playing of a animated gif for guizero
        widgets.

        :param Widget master:
            The widget which is displaying the animation.

        :param GUIZeroImage guizero_image:
            The image object which contains the animation.

        :param Callable update_image_callback:
            A function which should be called when the Image needs updating.
            The function will be called and passed a reference to the next
            Tk PhotoImage object in the animation.
        """
        self._master = master
        self._guizero_image = guizero_image
        self._update_image_callback = update_image_callback
        self._current_frame = 0
        self._running = False
        self.start()

    @property
    def running(self):
        """
        Returns `True` if the animation is running
        """
        return self._running

    def start(self):
        """
        Starts the animation.
        """
        if not self._running:
            self._running = True
            self._show_frame()

    def stop(self):
        """
        Stops the animation
        """
        self._running = False

    def _show_frame(self):
        if self.running:
            # get the frame
            frame_data = self._guizero_image.tk_frames[self._current_frame]
            frame = frame_data[0]
            delay = frame_data[1]

            # give it to the call back
            self._update_image_callback(frame)

            # increment the frame
            self._current_frame += 1
            if self._current_frame == len(self._guizero_image.tk_frames):
                self._current_frame = 0

            # call again after the delay
            self._master.after(delay, self._show_frame)


class TriggeredList(MutableSequence):

    def __init__(self, iterable=(), on_change=None):
        """
        A list which will call a callback when a value is changed.

        Used to hold a widgets grid reference.  
        """
        self._list = list(iterable)
        self._on_change = on_change

    def __getitem__(self, key):
        return self._list.__getitem__(key)

    def __setitem__(self, key, item):
        self._list.__setitem__(key, item)
        # trigger change handler
        self._changed()

    def __delitem__(self, key):
        self._list.__delitem__(key)
        # trigger change handler
        self._changed()

    def __len__(self):
        return self._list.__len__()

    def insert(self, index, item):
        self._list.insert(index, item)
        # trigger change handler
        self._changed()

    def _changed(self):
        if self._on_change is not None:
            self._on_change()

    def __str__(self):
        data = "["
        for item in self._list:
            data = data + str(item) + ", "
        data = data[:-2] + "]"
        return data


# Lambda-izer for making it easy to pass arguments with function calls
# without having to know what lambda does
def with_args( func_name, *args):
    return lambda: func_name(*args)

# Gets the number of args a function expects
def no_args_expected(func_name):
    args = getfullargspec(func_name).args
    if len(args) > 0:
        # if someone names the first arg of a class function to something
        # other than self, this will fail! or if they name the first argument
        # of a non class function to self this will fail!
        if args[0] == "self":
            return len(args) - 1
        else:
            return len(args)
    else:
        return 0

# Format errors in a pretty way
def error_format(message):
    print("------------------------------------------------------------")
    print("*** GUIZERO WARNING ***")
    print(message)
    print("------------------------------------------------------------")

# Raise error in a pretty way
def raise_error(message):
    error_message = "\n------------------------------------------------------------\n"
    error_message += "*** GUIZERO ERROR ***\n"
    error_message += message + "\n"
    error_message += "------------------------------------------------------------\n"
    raise GUIZeroException(error_message)

def deprecated(message):
    print("*** DEPRECATED: " + message)

def convert_color(color):
    """
    Converts a color from "color", (255, 255, 255) or "#ffffff" into a color tk 
    should understand.
    """
    if color is not None:

        # is the color a string
        if isinstance(color, str):
            # strip the color of white space
            color = color.strip()

            # if it starts with a # check it is a valid color
            if color[0] == "#":

                # check its format
                if len(color) != 7:
                    raise ValueError("{} is not a valid # color, it must be in the format #ffffff.".format(color))
                else:
                    # split the color into its hex values
                    hex_colors = (color[1:3], color[3:5], color[5:7])

                    # check hex values are between 00 and ff
                    for hex_color in hex_colors:
                        try:
                            int_color = int(hex_color, 16)
                        except:
                            raise ValueError("{} is not a valid value, it must be hex 00 - ff".format(hex_color))

                        if not (0 <= int_color <= 255):
                            raise ValueError("{} is not a valid color value, it must be 00 - ff".format(hex_color))

        # if the color is not a string, try and convert it
        else:
            # get the number of colors and check it is iterable
            try:
                no_of_colors = len(color)
            except:
                raise ValueError("A color must be a list or tuple of 3 values (red, green, blue)")

            if no_of_colors != 3:
                raise ValueError("A color must contain 3 values (red, green, blue)")

            # check the color values are between 0 and 255
            for c in color:
                if not (0 <= c <= 255):
                    raise ValueError("{} is not a valid color value, it must be 0 - 255")

            # convert to #ffffff format
            color = "#{:02x}{:02x}{:02x}".format(color[0], color[1], color[2])

    return color



from guizero import App, Waffle, PushButton

def clicked(x,y):
    waffle.set_pixel(x,y,"red")
    print("{}.{}".format(x,y))
    print(waffle.get_all())
    waffle.width = 5
    waffle.height = 5
    waffle.pixel_size = 50
    waffle.dotty = True
    waffle.pad = 25
    waffle.color = "green"

app = App()
waffle = Waffle(app, command = clicked, height = 15, width = 15)
waffle.bg=(255,255,0)
waffle.set_all("red")
waffle.set_pixel(0,0,"blue")
waffle.set_pixel(8,1,"#00ff00")
waffle.set_pixel(3,4,"blue")
waffle.pixel(1,2).dotty = True
print(waffle[8,1].color)

button = PushButton(app, command= waffle.reset)

app.display()



from guizero import *


def set_color(value):
	try:
		rgb = '#%02x%02x%02x' % (red.value, green.value, blue.value)
		coloured_text.value = rgb
		coloured_text.text_color = rgb
	except:
		print("Something went wrong with setting colours")


app = App()

title = Text(app, "Laura's awesome colour chooser", size=20, grid=[0,0])

sliderbox = Box(app, "grid")

r = Text(sliderbox, "R", grid=[1,1])
red = Slider(sliderbox, 0, 255, "horizontal", set_color, grid=[1,2])

g = Text(sliderbox, "G", grid=[2,1])
green = Slider(sliderbox, 0, 255, "horizontal", set_color, grid=[2,2])

b = Text(sliderbox, "B", grid=[3,1])
blue = Slider(sliderbox, 0, 255, "horizontal", set_color, grid=[3,2])


coloured_text = Text(app, "Coloured text", size=40, grid=[4,1])




app.display()



from tkinter import Canvas, BOTH, Frame
from . import utilities as utils
from .base import Widget
from .event import EventManager

class Waffle(Widget):

    def __init__(
        self,
        master,
        height=3,
        width=3,
        dim=20,
        pad=5,
        color="white",
        dotty=False,
        grid=None,
        align=None,
        command=None,
        visible=True,
        enabled=None,
        bg=None):

        # Create a tk Frame object within this object which will be the waffle
        tk = Frame(master.tk)

        # self._height = height       # How many pixels high
        # self._width = width         # How many pixels wide
        self._pixel_size = dim      # Size of one pixel
        self._pad = pad             # How much padding between pixels
        self._color = utils.convert_color(color)        # Start color of the whole waffle
        self._dotty = dotty         # A dotty waffle will display circles
        self._waffle_pixels = {}
        self._canvas = None

        super(Waffle, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        if bg is not None:
            self.bg = bg

        self.update_command(command)

        #override the event manager so it uses the canvas not the frame
        self._events = EventManager(self, self._canvas)

        # Bind the left mouse click to the canvas so we can click on the waffle
        self.events.set_event("<Waffle.ButtonPress-1>", "<ButtonPress-1>", self._clicked_on)


    # METHODS
    # -------------------------------------------

    # Internal use only
    def _create_waffle(self):
        if self._height == "fill" or self._width == "fill":
            utils.raise_error("{}\nCannot use 'fill' for width and height.".format(self.description))

        self._create_canvas()
        self._size_waffle()
        self._draw_waffle()

    def _create_canvas(self):
        # if the canvas exists, clear it and destroy it
        if self._canvas:
            self._canvas.delete("all")
            self._canvas.destroy()

        # size the canvas. Each element of the waffle has padding to top,
        # bottom, left and right, but the padding is shared between elements.
        # What this means in pratice is that elements at the edge have a whole
        # padding width between them and the edge of the canvas and a single
        # padding width between individual elements, so the padding between
        # elements is shared.
        self._c_height = (self._height * (self._pixel_size + self._pad)) + self._pad
        self._c_width = (self._width * (self._pixel_size + self._pad)) + self._pad

        # create the canvas and pack it into the waffle frame
        self._canvas = Canvas(self.tk, height=self._c_height, width=self._c_width, bd=0, highlightthickness=0)
        self._canvas.pack(fill=BOTH, expand=1)

        # rebind any events as they would have been lost when the canvas
        # was destroyed
        self.events.rebind_events(self._canvas)

        # fill the canvas background
        self._canvas.create_rectangle(0, 0, self._c_width, self._c_height, fill=self.bg, outline=self.bg)

    # sizes or resizes the waffle, maintaining the state of existing pixels
    def _size_waffle(self):
        # create new pixels
        new_waffle_pixels = {}
        currx = self._pad
        curry = self._pad

        for x in range(self._width):
            for y in range(self._height):
                if (x,y) in self._waffle_pixels.keys():
                    # if pixel already exist, reuse it and update the values
                    old_pixel = self._waffle_pixels[(x,y)]
                    new_waffle_pixels[x,y] = WafflePixel(
                        x, y,
                        self._canvas, currx, curry,
                        self._pixel_size,
                        old_pixel.dotty,
                        old_pixel.color)
                else:
                    # create a new celpixell
                    new_waffle_pixels[x,y] = WafflePixel(
                        x, y,
                        self._canvas, currx, curry,
                        self._pixel_size,
                        self._dotty,
                        self._color)

                curry += self._pixel_size + self._pad

            currx += self._pixel_size + self._pad
            curry = self._pad

        self._waffle_pixels = new_waffle_pixels

    def _draw_waffle(self):
        currx = self._pad
        curry = self._pad
        # draw the waffle from the waffle_cells
        for x in range(self._width):
            for y in range(self._height):
                cell = self._waffle_pixels[x,y]
                cell.draw()
                curry += self._pixel_size + self._pad

            currx += self._pixel_size + self._pad
            curry = self._pad

    # Sets the colour of the whole waffle
    def set_all(self, color):
        for x in range(self._width):
            for y in range(self._height):
                self._waffle_pixels[x,y].color = color

    # Sets a single pixel
    def set_pixel(self, x, y, color):
        if self.pixel(x, y):
            self._waffle_pixels[x,y].color = color

    # Returns the colour value of a pixel if set
    def get_pixel(self, x, y):
        if self.pixel(x, y):
            return self._waffle_pixels[x,y].color

    # Returns a 2D list of all colours in the waffle
    def get_all(self):
        all_pixels = []
        for y in range(self._height):
            row = []
            for x in range(self._width):
                row.append(self._waffle_pixels[x,y].color)
            all_pixels.append(row)
        return all_pixels

    # Internal use only
    # Detect x,y coords of where the user clicked
    def _clicked_on(self,e):
        # you can only click on the waffle if its enabled
        if self._enabled:
            canvas = e.tk_event.widget
            # Because the padding between waffle elements is shared we have to
            # determine which element is being clicked on if the user clicks on
            # the padding between elements and on the padding around the
            # elements at the edge of the waffle. The approach taken is that
            # half of the padding belongs to each element and padding at the
            # edges of the canvas are excluded
            x = int(canvas.canvasx(e.tk_event.x) - (self._pad / 2))
            y = int(canvas.canvasy(e.tk_event.y) - (self._pad / 2))
            if x >= 0 and y >= 0:
                pixel_x = int(x / (self._pixel_size + self._pad))
                pixel_y = int(y / (self._pixel_size + self._pad))
                if self._command and pixel_x < self._width and pixel_y < self._height:
                    args_expected = utils.no_args_expected(self._command)
                    if args_expected == 0:
                        self._command()
                    elif args_expected == 2:
                        self._command(pixel_x,pixel_y)
                    else:
                        utils.error_format("Waffle command function must accept either 0 or 2 arguments.\nThe current command has {} arguments.".format(args_expected))

    def update_command(self, command):
        if command is None:
            self._command = lambda: None
        else:
            self._command = command

    # Returns a WafflePixel object
    def pixel(self, x, y):
        if (x,y) in self._waffle_pixels.keys():
            _pixel = self._waffle_pixels[x,y]
        else:
            utils.error_format("Pixel x={} y={} is off the edge of the waffle".format(x, y))
            _pixel = None
        return _pixel

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value

    def disable(self):
        """Disable the widget."""
        self._enabled = False

    def enable(self):
        """Enable the widget."""
        self._enabled = True

    def resize(self, width, height):
        if self._width != width or self._height != height:
            self._width = width
            self._height = height
            self._create_waffle()

    @property
    def pixel_size(self):
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, value):
        if self._pixel_size != value:
            self._pixel_size = value
            self._create_waffle()

    @property
    def pad(self):
        return self._pad

    @pad.setter
    def pad(self, value):
        if self._pad != value:
            self._pad = value
            self._create_waffle()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        old_color = self._color
        self._color = utils.convert_color(value)
        for x in range(self._width):
            for y in range(self._height):
                if self._waffle_pixels[x,y].color == old_color:
                    self._waffle_pixels[x,y].color = self._color

    @property
    def dotty(self):
        return self._dotty

    @dotty.setter
    def dotty(self, value):
        self._dotty = value
        for x in range(self._width):
            for y in range(self._height):
                self._waffle_pixels[x,y].dotty = self._dotty

    # Get the background colour
    @property
    def bg(self):
        return super(Waffle, self.__class__).bg.fget(self)

    # Set the background colour
    @bg.setter
    def bg(self, value):
        # only change the color if we need too as it requires
        # redrawing the canvas
        if self.bg != value:
            value = utils.convert_color(value)
            super(Waffle, self.__class__).bg.fset(self, value)
            self._create_waffle()

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[Waffle] object with text ({}x{})".format(self._height, self._width)

    def reset(self):
        # reset all the colors and dottiness
        self.set_all(self._color)
        self.dotty = self._dotty

    def __getitem__(self, index):
        return self._waffle_pixels[index]

class WafflePixel():
    def __init__(self, x, y, canvas, canvas_x, canvas_y, size, dotty, color):
        self._x = x
        self._y = y
        self._canvas = canvas
        self._canvas_x = canvas_x
        self._canvas_y = canvas_y
        self._size = size
        self._dotty = dotty
        self._color = color
        self._drawn_object = None

    def draw(self):

        # if the object has been drawn before delete it
        if self._drawn_object:
            self._canvas.delete(self._drawn_object)

        if self._dotty == False:
            self._drawn_object = self._canvas.create_rectangle(
                self._canvas_x, self._canvas_y,
                self._canvas_x + self._size, self._canvas_y + self._size,
                fill=self._color)
        else:
            self._drawn_object = self._canvas.create_oval(
                self._canvas_x, self._canvas_y,
                self._canvas_x + self._size, self._canvas_y + self._size,
                fill=self._color)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def canvas_x(self):
        return self._canvas_x

    @property
    def canvas_y(self):
        return self._canvas_x

    @property
    def size(self):
        return self._size

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = utils.convert_color(value)
        self._canvas.itemconfig(self._drawn_object, fill=self._color)

    @property
    def dotty(self):
        return self._dotty

    @dotty.setter
    def dotty(self, value):
        self._dotty = value
        self.draw()



from tkinter import OptionMenu, StringVar, END, _setit
from . import utilities as utils
from .base import Base, TextWidget
from .tkmixins import ColorMixin, TextMixin


class ComboMenu(Base, ColorMixin, TextMixin):

    def __init__(self, tk):
        """
        Internal class for managing the little menu which pops up when the
        combo box is opened
        """
        super(ComboMenu, self).__init__(tk)


class Combo(TextWidget):

    def __init__(
        self,
        master,
        options=[],
        selected=None,
        command=None,
        grid=None,
        align=None,
        visible=True,
        enabled=None,
        width=None,
        height=None):
        """
        Creates a Combo

        :param Container master:
            The Container (App, Box, etc) the Combo will belong too.

        :param List option:
            A list of strings to populate the Combo, defaults to an empty list.

        :param string selected:
            The item in the Combo to select, defaults to `None`.

        :param Callable command:
            The callback function to call when the Combo changes,
            defaults to `None`.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        # Maintain a list of options (as strings, to avoid problems comparing)
        self._options = [str(x) for x in options]

        # Store currently selected item
        self._selected = StringVar()

        # Create a tk OptionMenu object within this object
        if len(self._options) == 0:
            tk = OptionMenu(master.tk, self._selected, None, command=self._command_callback)
        else:
            tk = OptionMenu(master.tk, self._selected, *self._options, command=self._command_callback)

        # Create the combo menu object
        self._combo_menu = ComboMenu(tk["menu"])

        super(Combo, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        # Remove the thick highlight when the bg is a different color
        self._set_tk_config("highlightthickness", 0)

        # Set the selected item
        if selected is None:
            # set the first option
            self._set_option_by_index(0)
        else:
            self._set_option(selected)

        self._default = selected

        # The command associated with this combo
        self.update_command(command)

    # PROPERTIES
    # ----------------------------------
    # The selected text value
    @property
    def value(self):
        """
        Sets or returns the option selected in a Combo.
        """
        return self._selected.get()

    @value.setter
    def value(self, value):
        value = str(value)
        if not self._set_option(value):
            utils.error_format("Tried to set " + self.description + " to option \"" + str(value) + "\", which does not exist" )

    @property
    def bg(self):
        """
        Sets or returns the background color of the widget.
        """
        return super(Combo, self.__class__).bg.fget(self)

    @bg.setter
    def bg(self, value):
        super(Combo, self.__class__).bg.fset(self, value)
        self._combo_menu.bg = value

    @property
    def text_color(self):
        """
        Sets or returns the text color used by the widget.
        """
        return super(Combo, self.__class__).text_color.fget(self)

    @text_color.setter
    def text_color(self, value):
        super(Combo, self.__class__).text_color.fset(self, value)
        self._combo_menu.text_color = value

    @property
    def text_size(self):
        """
        Sets or returns the text size used by the widget.
        """
        return super(Combo, self.__class__).text_size.fget(self)

    @text_size.setter
    def text_size(self, value):
        super(Combo, self.__class__).text_size.fset(self, value)
        self._combo_menu.text_size = value

    @property
    def font(self):
        """
        Sets or returns the font used by the widget.
        """
        return super(Combo, self.__class__).font.fget(self)

    @font.setter
    def font(self, value):
        super(Combo, self.__class__).font.fset(self, value)
        self._combo_menu.font = value

    @property
    def options(self):
        """
        Returns a list of options in the Combo
        """
        return self._options

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[Combo] object with options {}".format(self._options)

    # METHODS
    # -------------------------------------------

    # Resets the combo box to the original "selected" value from the constructor
    # (or the first value if no selected value was specified)
    def select_default(self):
        """
        Resets the combo box to the original "selected" value from the
        constructor (or the first value if no selected value was specified).
        """
        if self._default is None:
            if not self._set_option_by_index(0):
                utils.error_format(self.description + "\n" +
                "Unable to select default option as the Combo is empty")

        else:
            if not self._set_option(self._default):
                utils.error_format( self.description + "\n" +
                "Unable to select default option as it doesnt exist in the Combo")

    def append(self, option):
        """
        Appends a new `option` to the end of the Combo.

        :param string option:
            The option to append to the Combo.
        """
        self.insert(len(self._options), option)

    def insert(self, index, option):
        """
        Insert a new `option` in the Combo at `index`.

        :param int option:
            The index of where to insert the option.

        :param string option:
            The option to insert into to the Combo.
        """
        option = str(option)
        self._options.insert(index, option)
        # if this is the first option, set it.
        if len(self._options) == 1:
            self.value = option

        self._refresh_options()

    def remove(self, option):
        """
        Removes the first `option` from the Combo.

        Returns `True` if an item was removed.

        :param string option:
            The option to remove from the Combo.
        """
        if option in self._options:
            if len(self._options) == 1:
                # this is the last option in the list so clear it
                self.clear()
            else:
                self._options.remove(option)
                self._refresh_options()
                # have we just removed the selected option?
                # if so set it to the first option
                if option == self.value:
                    self._set_option(self._options[0])
            return True
        else:
            return False

    # Clear all options from the box
    def clear(self):
        """
        Clears all the options in a Combo
        """
        self._options = []
        self._combo_menu.tk.delete(0, END)
        self._selected.set("")

    def _refresh_options(self):
        # save which option was selected
        selected = self.value

        self._combo_menu.tk.delete(0, END)

        # Re-add all options
        # This uses an internal tk method _setit which is a bit naughty
        for item in self._options:
            self._combo_menu.tk.add_command(label=item, command=_setit(self._selected, item, self._command_callback))

        # set the option which was previously selected
        self._set_option(selected)

    def _set_option(self, value):
        """
        Sets a single option in the Combo, returning True if it was able too.
        """
        if len(self._options) > 0:
            if value in self._options:
                self._selected.set(value)
                return True
            else:
                return False
        else:
            return False

    def _set_option_by_index(self, index):
        """
        Sets a single option in the Combo by its index, returning True if it was able too.
        """
        if index < len(self._options):
            self._selected.set(self._options[index])
            return True
        else:
            return False

    def _command_callback(self, value):
        if self._command:
            args_expected = utils.no_args_expected(self._command)
            if args_expected == 0:
                self._command()
            elif args_expected == 1:
                self._command(value)
            else:
                utils.error_format("Combo command function must accept either 0 or 1 arguments.\nThe current command has {} arguments.".format(args_expected))

    def update_command(self, command):
        """
        Updates the callback command which is called when the Combo
        changes.

        Setting to `None` stops the callback.

        :param Callable command:
            The callback function to call, it can accept 0 or 1 parameters.

            If it accepts 1 parameter the `value` of the Combo will be
            passed.
        """
        if command is None:
            self._command = lambda: None
        else:
            self._command = command



from tkinter import Frame, StringVar
from . import utilities as utils
from .base import ContainerTextWidget
from .tkmixins import TextMixin
from .RadioButton import RadioButton
from .event import EventManager

class ButtonGroup(ContainerTextWidget):

    def __init__(
        self,
        master,
        options=[],
        selected=None,
        horizontal=False,
        command=None,
        grid=None,
        align=None,
        args=None,
        visible=True,
        enabled=None,
        width=None,
        height=None):
        """
        Creates a ButtonGroup

        :param Container master:
            The Container (App, Box, etc) the ButtonGroup will belong too.

        :param List option:
            A list of options to append to the ButtonGroup. If a 2D list is
            specified, the first element is the text, the second is the value,
            defaults to an empty list.

        :param string selected:
            The item in the ButtonGroup to select, defaults to `None`.

        :param string horizontal:
            If the ButtonGroup is to be displayed horizontally, defaults to
            `True`.

        :param Callable command:
            The callback function to call when the ButtonGroup changes,
            defaults to `None`.

        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the grid, defaults to None.

        :param callback args:
            A list of arguments to pass to the widgets `command`, defaults to
            `None`.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        """

        self._rbuttons = []   # List of RadioButton objects
        self._text_size = None
        self._font = None
        self._horizontal = horizontal

        # Create a Tk frame object to contain the RadioButton objects
        tk = Frame(master.tk)

        # Set (using StringVar set() method) the selected option **number**
        self._selected = StringVar(master=tk.winfo_toplevel())

        # ButtonGroup uses "grid" internally to sort the RadioButtons
        super(ButtonGroup, self).__init__(master, tk, "grid", grid, align, visible, enabled, width, height)

        # Loop through the list given and setup the options
        self._options = []
        for option in options:
            self._options.append(self._parse_option(option))

        self._refresh_options()

        # set the initial value
        if selected is None and len(self._options) > 0:
            self.value = self._options[0][1]
        else:
            self.value = selected

        # Add a command if there was one
        self.update_command(command, args)

        # override the event manager and associate the button group and the
        # radio buttons to it
        option_tks = [option.tk for option in self._rbuttons]
        self._events = EventManager(self, self.tk, *option_tks)

        # now the ButtonGroup is populate it, size it
        self.resize(width, height)

    def _parse_option(self, option):
        # If only a 1D  was provided, use the text value as a key
        if not isinstance(option, list):
            return [option, option]
        else:
            return [option[0], option[1]]

    def _refresh_options(self):
        # destroy any existing radio buttons
        for button in self._rbuttons:
            button.destroy()
        self._rbuttons = []

        gridx = 0
        gridy = 0

        for button in self._options:
            # Which way the buttons go
            if self._horizontal:
                gridx += 1
            else:
                gridy += 1

            # Create a radio button object
            rbutton = RadioButton(
                self,
                text=str(button[0]),
                value=str(button[1]),
                variable=self._selected,
                grid=[gridx, gridy],
                align="left",
                visible=self.visible,
                enabled=self.enabled)

            # Add this radio button to the internal list
            self._rbuttons.append(rbutton)

            # Set the callback
            rbutton.tk.config(command=self._command_callback)

    # PROPERTIES
    # -----------------------------------

    # Gets the selected value (1, 2, 3 etc.)
    @property
    def value(self):
        """
        Sets or returns the option selected in a ButtonGroup.
        """
        return (self._selected.get())

    # Sets which option is selected (if it doesn't exist, nothing is selected)
    @value.setter
    def value(self, value):
        self._selected.set(str(value))

    # Gets the text of the currently selected option
    @property
    def value_text(self):
        """
        Sets or returns the option selected in a ButtonGroup by its text value.
        """
        search = self._selected.get() # a string containing the selected option
        # This is a bit nasty - suggestions welcome
        for item in self._rbuttons:
            if item.value == search:
                return item.text
        return ""

    # Selects the option for the value_text provided
    @value_text.setter
    def value_text(self, value):
        for item in self._rbuttons:
            if item.text == value:
                self.value = item.value

    def resize(self, width, height):
        """
        Resizes the widget.

        :param int width:
            The width of the widget.

        :param int height:
            The height of the widget.
        """
        self._width = width
        self._height = height

        # update radio buttons width
        for item in self._rbuttons:
            item.width = width

        # update radio buttons height
        if len(self._rbuttons) > 0:
            # work out the height of a button
            button_height = height
            
            if isinstance(height, int):
                if height % len(self._rbuttons) != 0:
                    # if the height doesnt divide by the number of radio buttons give a warning
                    button_height = int(round(height / len(self._rbuttons)))
                    new_height = button_height * len(self._rbuttons)
                    utils.error_format("ButtonGroup height '{}' doesn't divide by the number of buttons '{}' setting height to '{}'.".format(height, len(self._rbuttons), new_height))
                else:
                    button_height = int(height / len(self._rbuttons))
            
            for item in self._rbuttons:
                item.height = button_height

        super(ButtonGroup, self).resize(width, height)

    @property
    def options(self):
        """
        Returns a list of options in the ButtonGroup
        """
        return self._options

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[ButtonGroup] object with selected option '{}'".format(self.value)

    # METHODS
    # -----------------------------------

    def append(self, option):
        """
        Appends a new `option` to the end of the ButtonGroup.

        :param string/List option:
            The option to append to the ButtonGroup. If a 2D list is specified,
            the first element is the text, the second is the value.
        """
        self._options.append(self._parse_option(option))
        self._refresh_options()
        self.resize(self._width, self._height)

    def insert(self, index, option):
        """
        Insert a new `option` in the ButtonGroup at `index`.

        :param int option:
            The index of where to insert the option.

        :param string/List option:
            The option to append to the ButtonGroup. If a 2D list is specified,
            the first element is the text, the second is the value.
        """
        self._options.insert(index, self._parse_option(option))
        self._refresh_options()
        self.resize(self._width, self._height)

    def remove(self, option):
        """
        Removes the first `option` from the ButtonGroup.

        Returns `True` if an item was removed.

        :param string option:
            The value of the option to remove from the ButtonGroup.
        """
        for existing_option in self._options:
            if existing_option[1] == option:
                self._options.remove(existing_option)
                self._refresh_options()
                return True
        return False

    def clear(self):
        """
        Clears all the options in a Combo
        """
        self._options = []
        self._refresh_options()
        self.value = ""

    # To help with debugging - return list of text/value pairs
    def get_group_as_list(self):
        return [[option.text, option.value] for option in self._rbuttons]

    def update_command(self, command, args=None):
        """
        Updates the callback command which is called when the ButtonGroup
        changes.

        Setting to `None` stops the callback.

        :param Callable command:
            The callback function to call.

        :param callback args:
            A list of arguments to pass to the widgets `command`, defaults to
            `None`.
        """
        if command is None:
            self._command = lambda: None
        else:
            if args is None:
                self._command = command
            else:
                self._command = utils.with_args(command, *args)

    def _command_callback(self):
        self._command()

 


from tkinter import Canvas, ALL
from tkinter.font import Font
from . import utilities as utils
from .base import Widget
from .event import EventManager

class Drawing(Widget):

    def __init__(
        self, 
        master, 
        width=100, 
        height=100, 
        grid=None, 
        align=None, 
        visible=True, 
        enabled=None):
        """
        Creates a Drawing

        :param Container master:
            The Container (App, Box, etc) the Drawing will belong too.

        :param int width:
            The starting width of the widget. Defaults to `None` and will auto
            size.

        :param int height:
            The starting height of the widget. Defaults to `None` and will auto
            size.
        
        :param List grid:
            Grid co-ordinates for the widget, required if the master layout
            is 'grid', defaults to `None`.

        :param string align:
            How to align the widget within the container, defaults to None.

        :param bool visible:
            If the widget should be visible, defaults to `True`.

        :param bool enabled:
            If the widget should be enabled, defaults to `None`. If `None`
            the value is inherited from the master.

        """

        # list to hold references to images, otherwise tk destroys them
        self._images = {}

        tk = Canvas(master.tk, height=100, width=100, bd=0, highlightthickness=0)

        super(Drawing, self).__init__(master, tk, grid, align, visible, enabled, width, height)

    def line(self, x1, y1, x2, y2, color="black", width=1):
        """
        Draws a line between 2 points

        :param int x1:
            The x position of the starting point.

        :param int y1:
            The y position of the starting point.

        :param int x2:
            The x position of the end point.

        :param int y2:
            The y position of the end point.

        :param str color:
            The color of the line. Defaults to `"black"`.

        :param int width:
            The width of the line. Defaults to `1`.

        :return:
            The id of the line.
        """
        return self.tk.create_line(
            x1, y1, x2, y2, 
            width = width,
            fill = "" if color is None else utils.convert_color(color)
            )

    def oval(self, x1, y1, x2, y2, color="black", outline=False, outline_color="black"):
        """
        Draws an oval between 2 points

        :param int x1:
            The x position of the starting point.

        :param int y1:
            The y position of the starting point.

        :param int x2:
            The x position of the end point.

        :param int y2:
            The y position of the end point.

        :param str color:
            The color of the shape. Defaults to `"black"`.

        :param int outline:
            `0` or `False` is no outline. `True` or value > 1 sets an outline. Defaults to `False`.

        :param str outline_color:
            The color of the outline. Defaults to `"black"`.

        :return:
            The id of the shape.
        """
        return self.tk.create_oval(
            x1, y1, x2, y2, 
            outline = utils.convert_color(outline_color) if outline else "",
            width = int(outline),
            fill = "" if color is None else utils.convert_color(color)
            )

    def rectangle(self, x1, y1, x2, y2, color="black", outline=False, outline_color="black"):
        """
        Draws a rectangle between 2 points

        :param int x1:
            The x position of the starting point.

        :param int y1:
            The y position of the starting point.

        :param int x2:
            The x position of the end point.

        :param int y2:
            The y position of the end point.

        :param str color:
            The color of the shape. Defaults to `"black"`.

        :param int outline:
            `0` or `False` is no outline. `True` or value > 1 sets an outline. Defaults to `False`.

        :param str outline_color:
            The color of the outline. Defaults to `"black"`.

        :return:
            The id of the shape.
        """
        return self.tk.create_rectangle(
            x1, y1, x2, y2, 
            outline = utils.convert_color(outline_color) if outline else "",
            width = int(outline),
            fill = "" if color is None else utils.convert_color(color)
            )

    def polygon(self, *coords, color="black", outline=False, outline_color="black"):
        """
        Draws a polygon from an list of co-ordinates

        :param int *coords:
            Pairs of x and y positions which make up the polygon.

        :param str color:
            The color of the shape. Defaults to `"black"`.

        :param int outline:
            `0` or `False` is no outline. `True` or value > 1 sets an outline. Defaults to `False`.

        :param str outline_color:
            The color of the outline. Defaults to `"black"`.

        :return:
            The id of the shape.
        """
        return self.tk.create_polygon(
            *coords, 
            outline = utils.convert_color(outline_color) if outline else "",
            width = int(outline),
            fill = "" if color is None else utils.convert_color(color)
            )

    def triangle(self, x1, y1, x2, y2, x3, y3, color="black", outline=False, outline_color="black"):
        """
        Draws a triangle between 3 points

        :param int x1:
            The x position of the starting point.

        :param int y1:
            The y position of the starting point.

        :param int x2:
            The x position of the middle point.

        :param int y2:
            The y position of the middle point.

        :param int x3:
            The x position of the end point.

        :param int y3:
            The y position of the end point.

        :param str color:
            The color of the shape. Defaults to `"black"`.

        :param int outline:
            `0` or `False` is no outline. `True` or value > 1 sets an outline. Defaults to `False`.

        :param str outline_color:
            The color of the outline. Defaults to `"black"`.

        :return:
            The id of the shape.
        """
        return self.polygon(
            x1, y1, x2, y2, x3, y3, 
            color=color, 
            outline=outline, 
            outline_color=outline_color)

    def image(self, x, y, image, width=None, height=None):
        """
        Inserts an image into the drawing, position by its top-left corner.
        
        :param int x:
            The x position to insert the image.

        :param int y:
            The y position to insert the image.

        :param str image:
            The file path or a PhotoImage or PIL.Image object.

        :param str width:
            The width to scale the image too, setting to `None` will use the
            actual width of the Image. Default to `None`.

        :param str height:
            The width to scale the image too, setting to `None` will use the 
            actual height of the Image. Default to `None`.

        :return:
            The id of the image.
        """
        # load the image and add to the dict (otherwise tk destroys the reference to them!)
        _image = utils.GUIZeroImage(image, width, height)
        id = self.tk.create_image(x, y, image=_image.tk_image, anchor="nw")
        self._images[id] = _image
        return id
    
    def text(self, x, y, text, color="black", font=None, size=None, max_width=None):
        """
        Inserts text into the drawing, position by its top-left corner.
        
        :param int x:
            The x position of the text.

        :param int y:
            The x position of the text.

        :param str color:
            The color of the text. Defaults to `"black"`.

        :param str font:
            The font to use. Defaults to `None` and will use the system
            default font.

        :param int size:
            The size of the text. Defaults to `None` and will use the system
            default font size.

        :param int max_width:
            Maximum line length. Lines longer than this value are wrapped. 
            Default is `None` (no wrapping).
        """
        # create the font
        if size is None:
            f = Font(self.tk, family=font)
        else:
            f = Font(self.tk, family=font, size=size)
        
        return self.tk.create_text(
            x, y, 
            text=text,
            fill = "" if color is None else utils.convert_color(color),
            font=f,
            width=max_width,
            anchor="nw")

    def delete(self, id):
        """
        Deletes an "object" (line, triangle, image, etc) from the drawing.

        :param int id:
            The id of the object.
        """
        if id in self._images.keys():
            del self._images[id]
        self.tk.delete(id)

    def clear(self):
        """
        Clears the drawing.
        """
        self._images.clear()
        self.tk.delete(ALL)



from guizero import App, Text, Picture
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    display_test,
    events_test,
    cascading_enable_test,
    cascading_properties_test,
    inheriting_properties_test,
    full_screen_test,
    add_tk_widget_test)

def test_default_values():
    a = App()
    assert a.title == "guizero"
    assert a.width == 500
    assert a.height == 500
    assert a.layout == "auto"
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(title = "foo", width = 666, height = 666, layout="grid")
    assert a.title == "foo"
    assert a.width == 666
    assert a.height == 666
    assert a.layout == "grid"
    a.destroy()

def test_getters_setters():
    a = App()
    a.title = "bar"
    assert a.title == "bar"
    a.bg = "red"
    assert a.bg == "red"
    a.height = 666
    assert a.height == 666
    a.width = 666
    assert a.width == 666
    a.destroy()

def test_update():
    a = App()
    # just testing it doesnt fail
    a.update()
    a.destroy()

def test_after_schedule():
    a = App()
    schedule_after_test(a, a)
    a.destroy()

def test_repeat_schedule():
    a = App()
    schedule_repeat_test(a, a)
    a.destroy()

def test_display():
    a = App()
    display_test(a)
    a.destroy()

def test_enable():
    a = App()
    t = Text(a)
    cascading_enable_test(a)
    a.destroy()

def test_events():
    a = App()
    events_test(a)
    a.destroy()

def test_cascading_properties():
    a = App()
    cascading_properties_test(a)
    a.destroy()

def test_inheriting_properties():
    a = App()
    inheriting_properties_test(a)
    a.destroy()

def test_full_screen():
    a = App()
    full_screen_test(a)
    a.destroy()

def test_add_tk_widget():
    a = App()
    add_tk_widget_test(a)
    a.destroy()


from guizero import *

# app = App(width=225, height=100)
app = App(width=225, height=130)

Box(app, width="fill", align="top", height=15)
Box(app, width="fill", align="bottom", height=15)

# box = Box(app, width=150, height=25)
# Text(box, text="Text in a Box")

# box = Box(app, border=True, width=150, height=25)
# Text(box, text="A Box with a border")

# choice = ButtonGroup(app, options=["cheese", "ham", "salad"])

# checkbox = CheckBox(app, text="salad ?")

# combo = Combo(app, options=["cheese", "ham", "salad"])

# drawing = Drawing(app, width=150, height=30)
# drawing.bg="green"
# drawing.rectangle(5, 5, 145, 25, color="blue")
# drawing.oval(7, 7, 23, 23, color="white")
# drawing.rectangle(25, 7, 39, 23, color="yellow")
# drawing.oval(120, 7, 142, 23, color="purple")
# drawing.text(50, 5, "Drawing", color="white")

# listbox = ListBox(app, items=["cheese", "ham", "salad"], height="fill")

# picture = Picture(app, image="guizero.png", height=100, width=200)

# def do_nothing():
#     print("button pressed")

# button = PushButton(app, command=do_nothing)

# slider = Slider(app)

# text = Text(app, text="Hello World")

# textbox = TextBox(app)

# waffle = Waffle(app)

window = Window(app, width=210, height=100)

app.display()   



from guizero import App, PushButton
def pressed():
    print("button was pressed")

def message(msg):
    print(msg)

app = App()
button = PushButton(app, command=pressed, text='Press me')
button_with_args = PushButton(app, command=message, text='Say hi', args=["hi"])
app.display()



from guizero import App, Box, Text, TextBox, PushButton

TEXT = """Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus. Donec quam felis, ultricies nec, pellentesque eu, pretium quis, sem. Nulla consequat massa quis enim. Donec pede justo, fringilla vel, aliquet nec, vulputate eget, arcu.

In enim justo, rhoncus ut, imperdiet a, venenatis vitae, justo. Nullam dictum felis eu pede mollis pretium. Integer tincidunt. Cras dapibus. Vivamus elementum semper nisi. Aenean vulputate eleifend tellus. Aenean leo ligula, porttitor eu, consequat vitae, eleifend ac, enim. Aliquam lorem ante, dapibus in, viverra quis, feugiat a, tellus.

Phasellus viverra nulla ut metus varius laoreet. Quisque rutrum. Aenean imperdiet. Etiam ultricies nisi vel augue. Curabitur ullamcorper ultricies nisi. Nam eget dui. Etiam rhoncus. Maecenas tempus, tellus eget condimentum rhoncus, sem quam semper libero, sit amet adipiscing sem neque sed ipsum. Nam quam nunc, blandit vel, luctus pulvinar, hendrerit id, lorem. Maecenas nec odio et ante tincidunt tempus. Donec vitae sapien ut libero venenatis faucibus. Nullam quis ante. Etiam sit amet orci eget eros faucibus tincidunt. Duis leo. Sed fringilla mauris sit amet nibh. Donec sodales sagittis magna. Sed consequat, leo eget bibendum sodales, augue velit cursus nunc."""

app = App()
Text(app, text="Some information")
TextBox(app, text=TEXT, width="fill", height="fill", multiline=True, scrollbar=True)
PushButton(app, text="Accept", align="bottom")
app.display()


from tkinter import Scale, HORIZONTAL, VERTICAL
from . import utilities as utils
from .base import TextWidget

class Slider(TextWidget):

    def __init__(
        self, 
        master, 
        start=0, 
        end=100, 
        horizontal=True, 
        command=None, 
        grid=None, 
        align=None, 
        visible=True, 
        enabled=None, 
        width=None, 
        height=None):

        self._start = start
        self._end = end

        # Set the direction
        self._horizontal = horizontal
        orient = HORIZONTAL if horizontal else VERTICAL

        # Create a tk Scale object within this object
        tk = Scale(master.tk, from_=start, to=end, orient=orient, command=self._command_callback)

        super(Slider, self).__init__(master, tk, grid, align, visible, enabled, width, height)

        self.update_command(command)

    # PROPERTIES
    # ----------------
    # Get/set the value
    @property
    def value(self):
        return (self.tk.get())

    @value.setter
    def value(self, value):
        self.tk.set(value)

    def resize(self, width, height):
        self._set_width(width)
        self._set_height(height)
        if width == "fill" or height == "fill":
            self.master.display_widgets()
    
    def _set_width(self, width):
        self._width = width
        if width != "fill":
            if self._horizontal:
                self._set_tk_config("length", width)
            else:
                self._set_tk_config("width", width)
    
    def _set_height(self, height):
        self._height = height
        if height != "fill":
            if self._horizontal:
                self._set_tk_config("width", height)
            else:
                self._set_tk_config("length", height)

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[Slider] object from {} to {}".format(self._start, self._end)

    # METHODS
    # ----------------
    # Calls the given function when the slider value is changed
    def _command_callback(self, value):
        if self._command:
            args_expected = utils.no_args_expected(self._command)
            if args_expected == 0:
                self._command()
            elif args_expected == 1:
                self._command(value)
            else:
                utils.error_format("Slider command function must accept either 0 or 1 arguments.\nThe current command has {} arguments.".format(args_expected))

    def update_command(self, command):
        if command is None:
            self._command = lambda: None
        else:
            self._command = command



from guizero import App, Box, Waffle, Slider, Text

def change_dim(slider):
    print("Changing size to {}x{} ".format(g_width.value, g_height.value))
    grid.resize(g_width.value, g_height.value)

app = App("Changing size")

grid = Waffle(app)

controls = Box(app, layout = "grid")

# Width
width_text = Text(controls, text="Width", grid=[0,0])
g_width = Slider(controls, start=0, end=20, command=change_dim, grid=[1,0])
g_width.value = 4

# Height
height_text = Text(controls, text="Height", grid=[0,1])
g_height = Slider(controls, start=0, end=20, command=change_dim, grid=[1,1])
g_height.value = 4

app.display()



from guizero import App, Waffle, PushButton

def change_pixel(x,y):
    if big_waffle.get_pixel(x,y) == 'white':
        big_waffle.set_pixel(x, y, "red")
    else:
        big_waffle.set_pixel(x, y, "white")

app = App("Waffle!", height=50*20, width=50*20)

big_waffle = Waffle(app, height=40, width=40, dim=10, pad=2, dotty=True, command=change_pixel)
enable_button = PushButton(app, big_waffle.enable, text="Enable")
disable_button = PushButton(app, big_waffle.disable, text="Disable")

app.display()



from guizero import App, Box, Text
from tkinter import Spinbox
from tkinter.ttk import Progressbar

app = App(title="Using tk widgets")

Text(app, text="Spinbox")

# add a spinbox widget to the app
sp = Spinbox(from_=0, to=10)
app.add_tk_widget(sp)

Text(app, text="and Progressbar")

box = Box(app, border=True)

# add a progress bar to the boc
pb = Progressbar(box.tk)
box.add_tk_widget(pb)
pb.start()

Text(app, text="in guizero")

app.display()


import os
from setuptools import setup, find_packages
from guizero import __name__, __package__, __version__, __author__

## This is a Python 3 package only
from sys import version_info
if version_info.major != 3:
    print("This package will only work with Python 3. \n"
          "If you already have Python 3 installed try 'pip3 install guizero'.")

__desc__ = 'Python module to allow learners to easily create GUIs'
__author_email__ = 'laura.sach@raspberrypi.org'
__license__ = 'BSD'
__url__ = 'https://github.com/lawsie/guizero'
__requires__ = []
__extras_require__={
        'images':  ["pillow>=4.3.0"]
    }
__python_requires__ = ">=3"
__keywords__ = [
    "GUI",
    "guizero",
    "interface",
]
__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Programming Language :: Python :: 3",
    "Topic :: Education",
    "Topic :: Software Development :: User Interfaces",
    "Topic :: Education",
    "Intended Audience :: Education",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
]
__long_description__ = """# guizero

guizero is designed to allow new learners to quickly and easily create GUIs for their programs.

There is comprehensive documentation at [lawsie.github.io/guizero](https://lawsie.github.io/guizero/)

## Install

If you can download and unzip a file, you can [install guizero](https://lawsie.github.io/guizero/#easy-install) - **no special permissions or administrator rights are required**.

If you have administrator rights and are connected to the internet, you can use [pip to quickly install guizero]((https://lawsie.github.io/guizero/#install-using-pip).

guizero only requires `tkinter` to be installed, which is included with a standard Python installation on all platforms except Linux.
A python module to allow learners to easily create GUIs. guizero is designed to be used by new learners.

## Use

guizero is simple to use, taking away much of the complexity of creating simple GUIs.

```python
from guizero import App, Text
app = App(title="Hello world")
message = Text(app, text="Welcome to the Hello world app!")
app.display()
```

"""

setup(
    name=__name__,
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    description=__desc__,
    long_description=__long_description__,
    long_description_content_type='text/markdown',
    license=__license__,
    keywords=__keywords__,
    url=__url__,
    packages=find_packages(),
    classifiers=__classifiers__,
    install_requires=__requires__,
    extras_require = __extras_require__,
    python_requires=__python_requires__,
)



from guizero import App, TextBox, PushButton, Box, Combo, CheckBox

def open_file():
    with open(file_name.value, "r") as f:
        editor.value = f.read()

def save_file():
    with open(file_name.value, "w") as f:
        f.write(editor.value)

def change_font():
    editor.font = font.value

def change_text_size():
    text_sizes = {"small": 6, "medium": 10, "large": 14}
    editor.text_size = text_sizes[size.value]
    # resize the widget because if the text is made bigger, it needs to be repacked,
    # and if you dont size it differently first tkinter wont repack it 
    editor.resize(1, 1)
    editor.resize("fill", "fill")

def change_theme():
    if dark_theme.value:
        app.bg = "black"
        app.text_color = "white"
    else:
        app.bg = None
        app.text_color = None
        
app = App(title="textzero")

file_controls = Box(app, align="top", width="fill", border=True)
file_name = TextBox(file_controls, text="text_file.txt", width=50, align="left")
save_button = PushButton(file_controls, text="Save", command=save_file, align="right")
open_button = PushButton(file_controls, text="Open", command=open_file, align="right")

editor = TextBox(app, multiline=True, height="fill", width="fill")

preferences_controls = Box(app, align="bottom", width="fill", border=True)
font = Combo(preferences_controls, options=["courier", "times new roman", "verdana"], align="left", command=change_font)
size = Combo(preferences_controls, options=["small", "medium", "large"], align="left", command=change_text_size)
dark_theme = CheckBox(preferences_controls, text="dark", align="left", command=change_theme)

app.display()


from guizero import App, Text, TextBox, Slider, Waffle, CheckBox, info

def intro():
    info("events", "this is all about events")

# when return is pressed back the test red and then turn it back to black
def key_pressed(e):

    def text_color_black():
        text_box.text_color = "black"

    # return pressed
    if ord(e.key) == 13:
        e.widget.text_color = "red"
        e.widget.after(1000, text_color_black)

def mouse_enters(e):
    e.widget.bg = "lightblue"

def mouse_leaves(e):
    e.widget.bg = "white"

def right_pressed():
    waffle.set_pixel(0,0,"green")

def right_released():
    waffle.set_pixel(0,0,"grey")

app = App()
text = Text(app, text="events")
slider = Slider(app)
text_box = TextBox(app)
check = CheckBox(app, "check")
waffle = Waffle(app)

# when clicked
text.when_clicked = intro

# when key pressed
text_box.when_key_pressed = key_pressed

# highlight widget when move over
check.when_mouse_enters = mouse_enters
check.when_mouse_leaves = mouse_leaves
slider.when_mouse_enters = mouse_enters
slider.when_mouse_leaves = mouse_leaves
text_box.when_mouse_enters = mouse_enters
text_box.when_mouse_leaves = mouse_leaves
waffle.when_mouse_enters = mouse_enters
waffle.when_mouse_leaves = mouse_leaves

# right button pressed and released
waffle.when_right_button_pressed = right_pressed
waffle.when_right_button_released = right_released

app.display()



from guizero import App, Text, PushButton

def count_down():
    timer.value = int(timer.value) - 1
    if timer.value == "0":
        # cancel counter
        timer.cancel(count_down)
        # reset counter
        timer.value = 10
        # setup repeat
        timer.repeat(1000, count_down)

app = App("timer")

timer = Text(app, text="10")
timer.repeat(1000, count_down)

app.display()


from guizero import App, Combo

def selected(value):
    print(value)

app = App()
combo = Combo(app, ["Nothing", "Something", "Everything"], command = selected, selected="Something")
combo.remove("Something")
combo2 = Combo(app)
combo2.append("hi")
combo2.append("bye")
app.display()


from guizero import App, PushButton, Text

def get_folder():
    file_name.value = app.select_folder()

app = App()

PushButton(app, command=get_folder, text="Get folder")
file_name = Text(app)

app.display()


from guizero import App, Text

number = 0
base = 2

def to_str(n,base):
    convert_string = "0123456789ABCDEF"
    if n < base:
        return convert_string[n]
    else:
        return to_str(n//base,base) + convert_string[n%base]

def counter():
    global number
    number = number + 1
    binary.value = str(to_str(number, base)).zfill(5)
    denary.value = str(number).zfill(5)

app = App()
denary = Text(app, text="00000", size=100)
binary = Text(app, text="00000", size=100)

denary.width = 10
binary.width = 10

denary.repeat(1000, counter)  # Schedule call to counter() every 1000ms
app.display()


import pytest
from threading import Event
from guizero import App, PushButton, system_config
from common_test import (
    schedule_after_test,
    schedule_repeat_test,
    destroy_test,
    enable_test,
    display_test,
    text_test,
    color_test,
    size_text_test,
    size_fill_test,
    events_test,
    cascaded_properties_test,
    inherited_properties_test,
    grid_layout_test,
    auto_layout_test
    )

def test_default_values():
    a = App()
    b = PushButton(a)
    assert b.master == a
    assert b.text == "Button"
    assert b.grid == None
    assert b.align == None
    assert a.description > ""
    a.destroy()

def test_alt_values():
    a = App(layout = "grid")
    b = PushButton(
        a,
        text = "foo",
        grid = [0,1],
        align = "top",
        width = 10,
        height = 11)

    assert b.text == "foo"
    assert b.grid[0] == 0
    assert b.grid[1] == 1
    assert b.align == "top"
    assert b.width == 10
    assert b.height == 11
    a.destroy()

def test_getters_setters():
    a = App()
    b = PushButton(a, text = "foo")
    assert b.text == "foo"
    b.text = "bar"
    assert b.text == "bar"
    a.destroy()

def test_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    b = PushButton(a, command = callback)
    assert not callback_event.is_set()
    b.tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_command_with_args():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "foo"
        callback_event.set()

    b = PushButton(a, command = callback, args = ["foo"])

    b.tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_update_command():
    a = App()

    callback_event = Event()
    def callback():
        callback_event.set()

    b = PushButton(a)

    b.tk.invoke()
    assert not callback_event.is_set()

    b.update_command(callback)
    b.tk.invoke()
    assert callback_event.is_set()
    callback_event.clear()

    b.update_command(None)
    b.tk.invoke()
    assert not callback_event.is_set()

    a.destroy()

def test_update_command_with_args():
    a = App()

    callback_event = Event()
    def callback(value):
        assert value == "foo"
        callback_event.set()

    b = PushButton(a)

    b.update_command(callback, ["foo"])
    b.tk.invoke()
    assert callback_event.is_set()

    a.destroy()

def test_toggle():
    a = App()
    b = PushButton(a)
    assert b.enabled
    b.toggle()
    assert not b.enabled
    b.toggle()
    assert b.enabled
    a.destroy()

def test_picture_gif():
    a = App()
    b = PushButton(a, image="../examples/guizero.gif")
    assert b.image == "../examples/guizero.gif"
    assert b._image.tk_image is not None
    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_picture_jpg():
    a = App()
    b = PushButton(a, image="../examples/guizero.jpg")
    assert b.image == "../examples/guizero.jpg"
    assert b._image.tk_image is not None
    assert b._image.pil_image is not None
    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_animated_picture():
    a = App()
    b = PushButton(a, image="../examples/guizero_flash.gif")
    assert b.image == "../examples/guizero_flash.gif"
    assert b._image.tk_image is not None
    assert b._image.pil_image is not None
    assert b._image.animation
    assert b._image_player.running
    a.destroy()

def test_picture_tkobject():
    from tkinter import PhotoImage

    a = App()
    photo_image = PhotoImage(file="../examples/guizero.gif")
    b = PushButton(a, image=photo_image)
    assert b.image == photo_image
    assert b._image.tk_image is not None
    a.destroy()

@pytest.mark.skipif(system_config.PIL_available == False,
                    reason="PIL not available")
def test_picture_pilobject():
    from PIL import Image

    a = App()
    pil_image = Image.open("../examples/guizero.gif")
    b = PushButton(a, image=pil_image)
    assert b.image == pil_image
    assert b._image.tk_image is not None
    assert b._image.pil_image is not None
    a.destroy()

def test_after_schedule():
    a = App()
    b = PushButton(a)
    schedule_after_test(a, b)
    a.destroy()

def test_repeat_schedule():
    a = App()
    b = PushButton(a)
    schedule_repeat_test(a, b)
    a.destroy()

def test_destroy():
    a = App()
    b = PushButton(a)
    destroy_test(b)
    a.destroy()

def test_enable():
    a = App()
    b = PushButton(a)
    enable_test(b)
    a.destroy()

def test_display():
    a = App()
    b = PushButton(a)
    display_test(b)
    a.destroy()

def test_text():
    a = App()
    b = PushButton(a)
    text_test(b)
    a.destroy()

def test_color():
    a = App()
    b = PushButton(a)
    color_test(b)
    a.destroy()

def test_size():
    a = App()
    b = PushButton(a)
    size_text_test(b)
    size_fill_test(b)
    a.destroy()

def test_events():
    a = App()
    p = PushButton(a)
    events_test(p)
    a.destroy()

def test_cascaded_properties():
    a = App()
    p = PushButton(a)
    cascaded_properties_test(a, p, True)
    a.destroy()

def test_inherited_properties():
    a = App()
    inherited_properties_test(a, lambda: PushButton(a), True)
    a.destroy()

def test_auto_layout():
    a = App()
    w = PushButton(a)
    auto_layout_test(w, None)
    a.destroy()

def test_grid_layout():
    a = App(layout="grid")
    
    w = PushButton(a, grid=[1,2])
    grid_layout_test(w, 1, 2, 1, 1, None)
    
    ws = PushButton(a, grid=[1,2,3,4])
    grid_layout_test(ws, 1, 2, 3, 4, None)

    wa = PushButton(a, grid=[1,2], align="top")
    grid_layout_test(wa, 1, 2, 1, 1, "top")
    
    a.destroy()


from guizero import App, ListBox, Text

def change_color(value):
    t.text_color = value

def update_text():
    if mlistbox.value is None:
        t.value = "Its a ListBox"
    else:
        t.value = "Its a " + " ".join(mlistbox.value) + " ListBox"

a = App()

t = Text(a, text="Its a ListBox", color="black")

listbox = ListBox(
    a,
    items=["red", "green", "blue", "yellow", "purple", "turquoise", "pink", "orange", "black", "brown", "cyan"],
    selected="black",
    command=change_color,
    scrollbar=True,
    width="fill",
    height=150)

mlistbox = ListBox(
    a,
    items=["really", "slightly", "brilliant", "interesting", "but", "and", "rubbish", "stupid"],
    multiselect=True,
    command=update_text,
    width="fill",
    height="fill")

a.display()



from guizero import App, PushButton, Text, CheckBox

def get_file():
    file_name.value = app.select_file(filetypes=[["All files", "*.*"], ["Text documents", "*.txt"]], save=save.value)

app = App()

save = CheckBox(app, text="save")
PushButton(app, command=get_file, text="Get file")
file_name = Text(app)

app.display()


from . import utilities as utils


class EventData():

    def __init__(self, widget, tk_event):
        """
        The EventData class represents a single event and is passed back to
        the callback
        """
        self._widget = widget
        self._tk_event = tk_event

    @property
    def widget(self):
        """
        The guizero widget which raised the event
        """
        return self._widget

    @property
    def tk_event(self):
        """
        The tkinter event which was returned by the event
        """
        return self._tk_event

    @property
    def key(self):
        """
        The key which was pressed and generated the event.
        """
        return self._tk_event.char

    @property
    def x(self):
        """
        The x position of the mouse relative to the upper left hand corner of
        the widget when the event occurred.
        """
        return self._tk_event.x

    @property
    def y(self):
        """
        The y position of the mouse relative to the upper left hand corner of
        the widget when the event occurred.
        """
        return self._tk_event.y

    @property
    def display_x(self):
        """
        The x position of the mouse relative to the upper left hand corner of
        the display when the event occurred.
        """
        return self._tk_event.x_root

    @property
    def display_y(self):
        """
        The y position of the mouse relative to the upper left hand corner of
        the display when the event occurred.
        """
        return self._tk_event.y_root


class EventCallback():

    def __init__(self, widget, tks, tk_event):
        """
        The EventCallback handles all the callbacks for a single tk event
        (e.g. <Button-1>) on a guizero widget.

        By using the EventCallback structure you can assign multiple callbacks
        to 1 tk event across multiple tk widgets.
        """
        self._widget = widget
        self._tks = tks
        self._tk_event = tk_event
        self._callbacks = {}
        self._func_ids = []

        # bind to the tk event
        for tk in self._tks:
            func_id = tk.bind(tk_event, self._event_callback)
            self._func_ids.append(func_id)

    def _event_callback(self, tk_event):
        # the tk event has fired, run all the callbacks associated to this event
        for ref in self._callbacks.copy():
            callback = self._callbacks[ref]
            args_expected = utils.no_args_expected(callback)

            if args_expected == 0:
                callback()
            elif args_expected == 1:
                callback(EventData(self._widget, tk_event))
            else:
                utils.error_format("An event callback function must accept either 0 or 1 arguments.\nThe current callback has {} arguments.".format(args_expected))

    def get_callback(self, ref):
        """
        Returns the callback for a ref (reference)
        """
        if ref in self._callbacks:
            return self._callbacks[ref]
        else:
            return None

    def set_callback(self, ref, callback):
        """
        Sets a callback for a ref (reference), setting to None will remove it
        """
        # if the callback already exists, remove it
        self.remove_callback(ref)

        # add it to the callbacks
        if callback is not None:
            self._callbacks[ref] = callback

    def remove_callback(self, ref):
        """
        Removes a callback for a ref (reference),
        """
        if ref in self._callbacks:
            del self._callbacks[ref]

    def rebind(self, tks):
        """
        Rebinds the tk event, only used if a widget has been destroyed
        and recreated.
        """
        self._tks = tks
        for tk in self._tks:
            tk.unbind_all(self._tk_event)
            func_id = tk.bind(self._tk_event, self._event_callback)
            self._func_ids.append(func_id)

    @property
    def widget(self):
        return self._widget

    @property
    def tk_event(self):
        return self._tk_event


class EventManager():

    def __init__(self, widget, *tks):
        """
        The EventManager handles all the events and callbacks for a guizero
        widget.

        Every event created must be given a reference, this reference
        is how events are managed internally within guizero.

        A guizero widget can contain many tk widgets, so all the tk objects
        for this guizero widget need to passed.
        """
        self._widget = widget
        self._tks = tks
        self._refs = {}
        self._event_callbacks = {}

    def get_event(self, ref):
        """
        Returns the event callback for a ref (reference)
        """
        # is this reference one which has been setup?
        if ref in self._refs:
            return self._refs[ref].get_callback(ref)
        else:
            return None

    def set_event(self, ref, tk_event, callback):
        """
        Sets a callback for this widget against a ref (reference) for a
        tk_event, setting the callback to None will remove it.
        """
        # has an EventCallback been created for this tk event
        if tk_event not in self._event_callbacks:
            self._event_callbacks[tk_event] = EventCallback(self._widget, self._tks, tk_event)

        # assign this ref to this event callback
        self._refs[ref] = self._event_callbacks[tk_event]

        # set up the callback
        self._refs[ref].set_callback(ref, callback)

    def remove_event(self, ref):
        """
        Removes an event for a ref (reference),
        """
        # is this reference one which has been setup?
        if ref in self._refs:
            self._refs[ref].remove_callback(ref)

    def rebind_events(self, *tks):
        """
        Rebinds all the tk events, only used if a tk widget has been destroyed
        and recreated.
        """
        for ref in self._refs:
            self._refs[ref].rebind(tks)



from tkinter import Radiobutton
from . import utilities as utils
from .base import TextWidget

# Only to be used as part of button group
# unless you want to externally create a controlling variable StringVar()
# Does NOT create an internal StringVar() as this is handled by the ButtonGroup class
class RadioButton(TextWidget):

    def __init__(self, master, text, value, variable, command=None, grid=None, align=None, visible=True, enabled=None):

        self._text = text
        self._value = value

        # `variable` is the externally passed StringVar keeping track of which
        # option was selected. This class should not be instantiated by a user
        # unless they know what they are doing.
        tk = Radiobutton(master.tk, text=self._text, value=self._value, variable=variable)

        super(RadioButton, self).__init__(master, tk, grid, align, visible, enabled, None, None)

    # PROPERTIES
    # -----------------------------------

    # The value of this button
    @property
    def value(self):
        return (self._value)

    @value.setter
    def value(self, value):
        self._value = str(value)
        self.tk.config(value=str(value))

    # The text from this button
    @property
    def text(self):
        return (self._text)

    @text.setter
    def text(self, text):
        self._text = str(text)
        self.tk.config(text=self._text)

    @property
    def description(self):
        """
        Returns the description for the widget.
        """
        return "[RadioButton] object with option={} value={}".format(self.text, self.value)